@@ -62,6 +62,30 @@
 (defvar ement-room-prism-minimum-contrast)
 (defvar ement-room-unread-only-counts-notifications)
 
+;;;; Function declarations
+
+;; Instead of using top-level `declare-function' forms (which can easily become obsolete
+;; if not kept with the code that needs them), this allows the use of `(declare (function
+;; ...))' forms in each function definition, so that if a function is moved or removed,
+;; the `declare-function' goes with it.
+
+;; TODO: Propose this upstream.
+
+(eval-and-compile
+  (defun ement--byte-run--declare-function (_name _args &rest values)
+    "Return a `declare-function' form with VALUES.
+Allows the use of a form like:
+
+  (declare (function FN FILE ...))
+
+inside of a function definition, effectively keeping its
+`declare-function' form inside the function definition, ensuring
+that stray such forms don't remain if the function is removed."
+    `(declare-function ,@values))
+
+  (cl-pushnew '(function ement--byte-run--declare-function) defun-declarations-alist :test #'equal)
+  (cl-pushnew '(function ement--byte-run--declare-function) macro-declarations-alist :test #'equal))
+
 ;;;; Compatibility
 
 ;; These workarounds should be removed when they aren't needed.
@@ -110,7 +134,7 @@ contrast colour with RGB as background and as foreground."
 ;;;;; Commands
 
 (cl-defun ement-create-room
-    (session &key name alias topic invite direct-p
+    (session &key name alias topic invite direct-p creation-content
              (then (lambda (data)
                      (message "Created new room: %s" (alist-get 'room_id data))))
              (visibility 'private))
@@ -118,7 +142,8 @@ contrast colour with RGB as background and as foreground."
 Then call function THEN with response data.  Optional string
 arguments are NAME, ALIAS, and TOPIC.  INVITE may be a list of
 user IDs to invite.  If DIRECT-P, set the \"is_direct\" flag in
-the request."
+the request.  CREATION-CONTENT may be an alist of extra keys to
+include with the request (see Matrix spec)."
   ;; TODO: Document other arguments.
   ;; SPEC: 10.1.1.
   (declare (indent defun))
@@ -126,7 +151,7 @@ the request."
 		     :name (read-string "New room name: ")
 		     :alias (read-string "New room alias (e.g. \"foo\" for \"#foo:matrix.org\"): ")
 		     :topic (read-string "New room topic: ")
-		     :visibility (completing-read "New room type: " '(private public))))
+		     :visibility (completing-read "New room visibility: " '(private public))))
   (cl-labels ((given-p
 	       (var) (and var (not (string-empty-p var)))))
     (pcase-let* ((endpoint "createRoom")
@@ -141,18 +166,38 @@ the request."
 			 (when invite
 			   (push (cons "invite" invite) it))
 			 (when direct-p
-			   (push (cons "is_direct" t) it)))))
+			   (push (cons "is_direct" t) it))
+                         (when creation-content
+                           (push (cons "creation_content" creation-content) it)))))
       (ement-api session endpoint :method 'post :data (json-encode data)
         :then then))))
 
+(cl-defun ement-create-space
+    (session &key name alias topic
+             (then (lambda (data)
+                     (message "Created new space: %s" (alist-get 'room_id data))))
+             (visibility 'private))
+  "Create new space on SESSION.
+Then call function THEN with response data.  Optional string
+arguments are NAME, ALIAS, and TOPIC."
+  (declare (indent defun))
+  (interactive (list (ement-complete-session)
+		     :name (read-string "New space name: ")
+		     :alias (read-string "New space alias (e.g. \"foo\" for \"#foo:matrix.org\"): ")
+		     :topic (read-string "New space topic: ")
+		     :visibility (completing-read "New space visibility: " '(private public))))
+  (ement-create-room session :name name :alias alias :topic topic :visibility visibility
+    :creation-content (ement-alist "type" "m.space") :then then))
+
 (defun ement-room-leave (room session &optional force-p)
   "Leave ROOM on SESSION.
 If FORCE-P, leave without prompting.  ROOM may be an `ement-room'
 struct, or a room ID or alias string."
   ;; TODO: Rename `room' argument to `room-or-id'.
-  (interactive (ement-complete-room :session (ement-complete-session)
-                 :prompt "Leave room: "))
-  (cl-assert room) (cl-assert session)
+  (interactive
+   (ement-with-room-and-session
+     :prompt-form (ement-complete-room :prompt "Leave room: ")
+     (list ement-room ement-session)))
   (cl-etypecase room
     (ement-room)
     (string (setf room (ement-afirst (or (equal room (ement-room-canonical-alias it))
@@ -176,7 +221,7 @@ struct, or a room ID or alias string."
                                          (kill-buffer buffer))))))
                     (setf (symbol-function leave-fn-symbol) leave-fn)
                     (add-hook 'ement-sync-callback-hook leave-fn-symbol)))
-                (message "Left room: %s" (ement--format-room room)))
+                (ement-message "Left room: %s" (ement--format-room room)))
         :else (lambda (plz-error)
                 (pcase-let* (((cl-struct plz-error response) plz-error)
                              ((cl-struct plz-response status body) response)
@@ -190,8 +235,10 @@ struct, or a room ID or alias string."
   "Forget ROOM on SESSION.
 If FORCE-P (interactively, with prefix), prompt to leave the room
 when necessary, and forget the room without prompting."
-  (interactive (ement-complete-room :session (ement-complete-session)
-                 :prompt "Forget room: "))
+  (interactive
+   (ement-with-room-and-session
+     :prompt-form (ement-complete-room :prompt "Forget room: ")
+     (list ement-room ement-session current-prefix-arg)))
   (pcase-let* (((cl-struct ement-room id display-name status) room)
                (endpoint (format "rooms/%s/forget" (url-hexify-string id))))
     (pcase status
@@ -220,7 +267,7 @@ when necessary, and forget the room without prompting."
                           (setf (ement-session-rooms session)
                                 (cl-remove room (ement-session-rooms session)))
                           ;; TODO: Indicate forgotten in footer in room buffer.
-                          (message "Room \"%s\" (%s) forgotten." display-name id))))))))
+                          (ement-message "Forgot room: %s." (ement--format-room room)))))))))
 
 (defun ement-ignore-user (user-id session &optional unignore-p)
   "Ignore USER-ID on SESSION.
@@ -244,13 +291,13 @@ If UNIGNORE-P (interactively, with prefix), un-ignore USER."
               (message "Ement: User %s %s." user-id (if unignore-p "unignored" "ignored"))))))
 
 (defun ement-invite-user (user-id room session)
-  "Invite USER-ID to ROOM on SESSION."
+  "Invite USER-ID to ROOM on SESSION.
+Interactively, with prefix, prompt for room and session,
+otherwise use current room."
   ;; SPEC: 10.4.2.1.
   (interactive
-   (let* ((session (ement-complete-session))
-          (user-id (ement-complete-user-id))
-          (room (car (ement-complete-room :session session))))
-     (list user-id room session)))
+   (ement-with-room-and-session
+     (list (ement-complete-user-id) ement-room ement-session)))
   (pcase-let* ((endpoint (format "rooms/%s/invite"
                                  (url-hexify-string (ement-room-id room))))
                (data (ement-alist "user_id" user-id) ))
@@ -263,9 +310,13 @@ If UNIGNORE-P (interactively, with prefix), un-ignore USER."
 
 (defun ement-list-members (room session bufferp)
   "Show members of ROOM on SESSION.
-If BUFFERP (interactively, with prefix), or if there are many
-members, show in a new buffer; otherwise show in echo area."
-  (interactive (list ement-room ement-session current-prefix-arg))
+Interactively, with prefix, prompt for room and session,
+otherwise use current room.  If BUFFERP (interactively, with
+prefix), or if there are many members, show in a new buffer;
+otherwise show in echo area."
+  (interactive
+   (ement-with-room-and-session
+     (list ement-room ement-session current-prefix-arg)))
   (pcase-let* (((cl-struct ement-room members (local (map fetched-members-p))) room)
                (list-members
                 (lambda (&optional _)
@@ -342,30 +393,38 @@ new one automatically if necessary."
                 (message "Room \"%s\" created for user %s.  Sending message..."
 	                 room-id user-id))))))
 
-(defun ement-tag-room (tag room session &optional delete)
-  "Add TAG to ROOM on SESSION.
-If DELETE (interactively, with prefix), delete it."
+(defun ement-tag-room (tag room session)
+  "Toggle TAG for ROOM on SESSION."
   (interactive
-   (pcase-let* ((`(,room ,session) (or (when (bound-and-true-p ement-room)
-                                         (list ement-room ement-session))
-                                       (ement-complete-room)))
-                (prompt (if current-prefix-arg "Delete tag: " "Add tag: "))
-                (default-tags (ement-alist "Favourite" "m.favourite"
-                                           "Low-priority" "m.lowpriority"))
-                (input (completing-read prompt default-tags))
-                (tag (alist-get input default-tags (concat "u." input) nil #'string=)))
-     (list tag room session current-prefix-arg)))
+   (ement-with-room-and-session
+     (let* ((prompt (format "Toggle tag (%s): " (ement--format-room ement-room)))
+            (default-tags
+              (ement-alist (propertize "Favourite"
+                                       'face (when (ement--room-tagged-p "m.favourite" ement-room)
+                                               'transient-value))
+                           "m.favourite"
+                           (propertize "Low-priority"
+                                       'face (when (ement--room-tagged-p "m.lowpriority" ement-room)
+                                               'transient-value))
+                           "m.lowpriority"))
+            (input (completing-read prompt default-tags))
+            (tag (alist-get input default-tags (concat "u." input) nil #'string=)))
+       (list tag ement-room ement-session))))
   (pcase-let* (((cl-struct ement-session user) session)
                ((cl-struct ement-user (id user-id)) user)
                ((cl-struct ement-room (id room-id)) room)
                (endpoint (format "user/%s/rooms/%s/tags/%s"
                                  (url-hexify-string user-id) (url-hexify-string room-id) (url-hexify-string tag)))
-               (method (if delete 'delete 'put)))
+               (method (if (ement--room-tagged-p tag room) 'delete 'put)))
     ;; TODO: "order".
     ;; FIXME: Removing a tag on a left room doesn't seem to work (e.g. to unfavorite a room after leaving it, but not forgetting it).
-    (ement-api session endpoint :version "v3" :method method :data (unless delete "{}")
-      :then (lambda (data)
-              (ement-debug "Changed tag on room" method tag data room)))))
+    (ement-api session endpoint :version "v3" :method method :data (pcase method ('put "{}"))
+      :then (lambda (_)
+              (ement-message "%s tag %S on %s"
+                             (pcase method
+                               ('delete "Removed")
+                               ('put "Added"))
+                             tag (ement--format-room room)) ))))
 
 (defun ement-set-display-name (display-name session)
   "Set DISPLAY-NAME for user on SESSION.
@@ -386,16 +445,16 @@ Sets global displayname."
 
 (defun ement-room-set-display-name (display-name room session)
   "Set DISPLAY-NAME for user in ROOM on SESSION.
-Sets the name only in ROOM, not globally."
+Interactively, with prefix, prompt for room and session,
+otherwise use current room.  Sets the name only in ROOM, not
+globally."
   (interactive
-   (pcase-let* ((`(,room ,session) (or (when (bound-and-true-p ement-room)
-                                         (list ement-room ement-session))
-                                       (ement-complete-room)))
-                (prompt (format "Set display-name in %S to: "
-                                (ement--format-room room)))
-                (display-name (read-string prompt nil nil
-                                           (ement-user-displayname (ement-session-user session)))))
-     (list display-name room session)))
+   (ement-with-room-and-session
+     (let* ((prompt (format "Set display-name in %S to: "
+                            (ement--format-room ement-room)))
+            (display-name (read-string prompt nil nil
+                                       (ement-user-displayname (ement-session-user ement-session)))))
+       (list display-name ement-room ement-session))))
   ;; NOTE: This does not seem to be documented in the spec, so we imitate the
   ;; "/myroomnick" command in SlashCommands.tsx from matrix-react-sdk.
   (pcase-let* (((cl-struct ement-room state) room)
@@ -426,9 +485,10 @@ Sets the name only in ROOM, not globally."
   "Ement-Describe-Room" "Major mode for `ement-describe-room' buffers.")
 
 (defun ement-describe-room (room session)
-  "Describe ROOM on SESSION."
-  (interactive (pcase-let ((`(,room ,session) (ement-complete-room :session ement-session)))
-                 (list room session)))
+  "Describe ROOM on SESSION.
+Interactively, with prefix, prompt for room and session,
+otherwise use current room."
+  (interactive (ement-with-room-and-session (list ement-room ement-session)))
   (cl-labels ((heading (string)
                        (propertize (or string "") 'face 'font-lock-builtin-face))
               (id (string)
@@ -558,8 +618,9 @@ Returns one of nil (meaning default rules are used), `all-loud',
 
 (defun ement-room-set-notification-state (state room session)
   "Set notification STATE for ROOM on SESSION.
-STATE may be nil to set the rules to default, `all',
-`mentions-and-keywords', or `none'."
+Interactively, with prefix, prompt for room and session,
+otherwise use current room.  STATE may be nil to set the rules to
+default, `all', `mentions-and-keywords', or `none'."
   ;; This merely attempts to reproduce the behavior of Element's simple notification
   ;; options.  It does not attempt to offer all of the features defined in the spec.  And,
   ;; yes, it is rather awkward, having to sometimes* make multiple requests of different
@@ -576,17 +637,15 @@ STATE may be nil to set the rules to default, `all',
 
   ;; TODO: Support `all-loud' ("all_messages_loud").
   (interactive
-   (pcase-let* ((`(,room ,session) (or (when (bound-and-true-p ement-room)
-                                         (list ement-room ement-session))
-                                       (ement-complete-room)))
-                (prompt (format "Set notification rules for %s: " (ement--format-room room)))
-                (available-states (ement-alist "Default" nil
-                                               "All messages" 'all
-                                               "Mentions and keywords" 'mentions-and-keywords
-                                               "None" 'none))
-                (selected-rule (completing-read prompt (mapcar #'car available-states) nil t))
-                (state (alist-get selected-rule available-states nil nil #'equal)))
-     (list state room session)))
+   (ement-with-room-and-session
+     (let* ((prompt (format "Set notification rules for %s: " (ement--format-room ement-room)))
+            (available-states (ement-alist "Default" nil
+                                           "All messages" 'all
+                                           "Mentions and keywords" 'mentions-and-keywords
+                                           "None" 'none))
+            (selected-rule (completing-read prompt (mapcar #'car available-states) nil t))
+            (state (alist-get selected-rule available-states nil nil #'equal)))
+       (list state ement-room ement-session))))
   (cl-labels ((set-rule (kind rule queue message-fn)
                         (pcase-let* (((cl-struct ement-room (id room-id)) room)
                                      (rule-id (url-hexify-string room-id))
@@ -774,10 +833,56 @@ USER is an `ement-user' struct."
 ;; These functions aren't expected to be called by code in other packages (but if that
 ;; were necessary, they could be renamed accordingly).
 
+;; (defun ement--room-routing (room)
+;;   "Return a list of servers to route to ROOM through."
+;;   ;; See <https://spec.matrix.org/v1.2/appendices/#routing>.
+;;   ;; FIXME: Ensure highest power level user is at least level 50.
+;;   ;; FIXME: Ignore servers blocked due to server ACLs.
+;;   ;; FIXME: Ignore servers which are IP addresses.
+;;   (cl-labels ((most-powerful-user-in
+;;                (room))
+;;               (servers-by-population-in
+;;                (room))
+;;               (server-of (user)))
+;;     (let (first-server-by-power-level)
+;;       (delete-dups
+;;        (remq nil
+;;              (list
+;;               ;; 1.
+;;               (or (when-let ((user (most-powerful-user-in room)))
+;;                     (setf first-server-by-power-level t)
+;;                     (server-of user))
+;;                   (car (servers-by-population-in room)))
+;;               ;; 2.
+;;               (if first-server-by-power-level
+;;                   (car (servers-by-population-in room))
+;;                 (cl-second (servers-by-population-in room)))
+;;               ;; 3.
+;;               (cl-third (servers-by-population-in room))))))))
+
 (defun ement--room-space-p (room)
   "Return non-nil if ROOM is a space."
   (equal "m.space" (ement-room-type room)))
 
+(defun ement--room-in-space-p (room space)
+  "Return non-nil if ROOM is in SPACE on SESSION."
+  ;; We could use `ement---room-spaces', but since that returns rooms by looking them up
+  ;; by ID in the session's rooms list, this is more efficient.
+  (pcase-let* (((cl-struct ement-room (id parent-id) (local (map children))) space)
+               ((cl-struct ement-room (id child-id) (local (map parents))) room))
+    (or (member parent-id parents)
+        (member child-id children))))
+
+(defun ement--room-spaces (room session)
+  "Return list of ROOM's parent spaces on SESSION."
+  ;; NOTE: This only looks in the room's parents list; it doesn't look in every space's children
+  ;; list.  This should be good enough, assuming we add to the lists correctly elsewhere.
+  (pcase-let* (((cl-struct ement-session rooms) session)
+               ((cl-struct ement-room (local (map parents))) room))
+    (cl-remove-if-not (lambda (session-room-id)
+                        (member session-room-id parents))
+                      rooms :key #'ement-room-id)))
+
 (cl-defun ement--prism-color (string &key (contrast-with (face-background 'default nil 'default)))
   "Return a computed color for STRING.
 The color is adjusted to have sufficient contrast with the color
@@ -900,21 +1005,13 @@ period, anywhere in the body."
     (pcase-let* (((cl-struct ement-room members) room)
                  (regexp (rx (or bos bow (1+ blank))
                              (or (seq (group
-                                       ;; Group 1: full MXID or @-prefixed displayname.
+                                       ;; Group 1: full @-prefixed MXID.
                                        "@" (group
-                                            ;; Group 2: MXID username or displayname.
-                                            ;; NOTE: We special-case the question mark,
-                                            ;; period, and comma so they can be used after
-                                            ;; a displayname or MXID in a sentence (if
-                                            ;; they are present in a displayname, too
-                                            ;; bad).
-                                            (1+ (not (any blank "?.,")))))
-                                      ;; NOTE: Including punctuation in this terminator
-                                      ;; means that wild displaynames with non-word
-                                      ;; characters might not get matched, but it's
-                                      ;; necessary so that a mention can be like "What
-                                      ;; about @foobar?"
-                                      (or eow eos (syntax punctuation) (seq ":" (1+ blank))))
+                                            ;; Group 2: displayname.  (NOTE: Does not work
+                                            ;; with displaynames containing spaces.)
+                                            (1+ (seq (optional ".") alnum)))
+                                       (optional ":" (1+ (seq (optional ".") alnum))))
+                                      (or ":" eow eos (syntax punctuation)))
                                  (seq (group
                                        ;; Group 3: MXID username or displayname.
                                        (1+ (not blank)))
@@ -927,14 +1024,9 @@ period, anywhere in the body."
                         (setf replace-group 1)
                         (format template (match-string 1 body)
                                 (ement--xml-escape-string (ement--user-displayname-in room member))))
-                      (when-let* ((name (or (when (match-string 1 body)
-                                              ;; Found @-prefixed displayname: replace the
-                                              ;; whole thing, including the @, but only
-                                              ;; use the displayname for looking up the user.
+                      (when-let* ((name (or (when (match-string 2 body)
                                               (setf replace-group 1)
                                               (match-string 2 body))
-                                            (prog1 (match-string 2 body)
-                                              (setf replace-group 2))
                                             (prog1 (match-string 3 body)
                                               (setf replace-group 3))))
                                   (members (members-having-displayname name members))
@@ -1021,10 +1113,10 @@ e.g. `ement-room-send-org-filter')."
                             (push (cons "formatted_body" formatted-body) it)
                             (push (cons "format" "org.matrix.custom.html") it))))
                (then (or then #'ignore)))
-    (when replying-to-event
-      (setf content (ement--add-reply content replying-to-event room)))
     (when filter
       (setf content (funcall filter content room)))
+    (when replying-to-event
+      (setf content (ement--add-reply content replying-to-event room)))
     (ement-api session endpoint :method 'put :data (json-encode content)
       :then (apply-partially then :room room :session session
                              ;; Data is added when calling back.
